/*-*- Mode: C -*-
 vi:si:et:sw=4:sts=4:ts=4 
 */

#include <Python.h>
#include "silcincludes.h"
#include "silcclient.h"

#define TRUE  1
#define FALSE 0

typedef struct {
    PyObject_HEAD
    SilcClient client;
    SilcClientConnection conn;
    bool connected;
    SilcHashTable channels;
} silcclientobject;	

typedef struct {
    char *name;
    PyObject *callback;
} callbackobject;

staticforward PyTypeObject SilcClientObjectType;
static PyObject *SilcClientError = NULL;

#define SILC_SAY               0
#define SILC_CHANNEL_MESSAGE   1
#define SILC_PRIVATE_MESSAGE   2
#define SILC_NOTIFY            3
#define SILC_COMMAND           4
#define SILC_COMMAND_REPLY     5
#define SILC_CONNECTED         6
#define SILC_DISCONNECTED      7
#define SILC_GET_AUTH_METHOD   8
#define SILC_VERIFY_PUBLIC_KEY 9
#define SILC_ASK_PASSPHRASE    10
#define SILC_FAILURE           11
#define SILC_KEY_AGREEMENT     12
#define SILC_FTP               13
#define SILC_DETACH            14


static callbackobject callbacks[] = {
    {"say",               NULL},
    {"channel_message",   NULL},
    {"private_message",   NULL},
    {"notify",            NULL},
    {"command",           NULL},
    {"command_reply",     NULL},
    {"connected",         NULL},
    {"disconnected",      NULL},
    {"get_auth_method",   NULL},
    {"verify_public_key", NULL},
    {"ask_passphrase",    NULL},
    {"failure",           NULL},
    {"key_agreement",     NULL},
    {"ftp",               NULL},
    {"detach",            NULL},
    {NULL,                NULL}
};

bool hash_compare_str(void *key1, void *key2, void *user_context)
{
    int l1;
    l1 = strlen((char *)key1);
    if (strncmp((char *)key1, (char *)key2, l1) == 0)
        return TRUE;
    else
        return FALSE;
}

static void silc_get_umode_string(SilcUInt32 mode, char *buf,
				  SilcUInt32 buf_size)
{
  if ((mode & SILC_UMODE_SERVER_OPERATOR) ||
      (mode & SILC_UMODE_ROUTER_OPERATOR)) {
    strcat(buf, (mode & SILC_UMODE_SERVER_OPERATOR) ?
	   "[server operator]" :
	   (mode & SILC_UMODE_ROUTER_OPERATOR) ?
	   "[SILC operator]" : "[unknown mode]");
  }
  if (mode & SILC_UMODE_GONE)
    strcat(buf, " [away]");
  if (mode & SILC_UMODE_INDISPOSED)
    strcat(buf, " [indisposed]");
  if (mode & SILC_UMODE_BUSY)
    strcat(buf, " [busy]");
  if (mode & SILC_UMODE_PAGE)
    strcat(buf, " [page to reach]");
  if (mode & SILC_UMODE_HYPER)
    strcat(buf, " [hyper active]");
  if (mode & SILC_UMODE_ROBOT)
    strcat(buf, " [robot]");
  if (mode & SILC_UMODE_ANONYMOUS)
    strcat(buf, " [anonymous]");
  if (mode & SILC_UMODE_BLOCK_PRIVMSG)
    strcat(buf, " [blocks private messages]");
  if (mode & SILC_UMODE_DETACHED)
    strcat(buf, " [detached]");
  if (mode & SILC_UMODE_REJECT_WATCHING)
    strcat(buf, " [rejects watching]");
  if (mode & SILC_UMODE_BLOCK_INVITE)
    strcat(buf, " [blocks invites]");
}

static void stub_silc_say(SilcClient client, SilcClientConnection conn,
                          SilcClientMessageType type, char *msg, ...)
{
    char *str;
    va_list va;
    va_start(va, msg);
    vasprintf(&str, msg, va);
    va_end(va);
    PyObject *arglist = Py_BuildValue("(s)", str);

    if (callbacks[SILC_SAY].callback != NULL)
        PyEval_CallObject(callbacks[SILC_SAY].callback, arglist);
}

static void stub_silc_channel_message(SilcClient client, SilcClientConnection conn,
                                      SilcClientEntry sender, SilcChannelEntry channel,
                                      SilcMessagePayload payload,
                                      SilcChannelPrivateKey key,
                                      SilcMessageFlags flags, const unsigned char *message,
                                      SilcUInt32 message_len)
{
    PyObject *arglist = Py_BuildValue("ss", sender->nickname, message);
    if (callbacks[SILC_CHANNEL_MESSAGE].callback != NULL)
        PyEval_CallObject(callbacks[SILC_CHANNEL_MESSAGE].callback, arglist);
}

static void stub_silc_private_message(SilcClient client, SilcClientConnection conn,
                                      SilcClientEntry sender, SilcMessagePayload payload,
                                      SilcMessageFlags flags,
                                      const unsigned char *message,
                                      SilcUInt32 message_len)
{
    PyObject *arglist = Py_BuildValue("ss", sender->nickname, message);
    if (callbacks[SILC_PRIVATE_MESSAGE].callback != NULL)
        PyEval_CallObject(callbacks[SILC_PRIVATE_MESSAGE].callback, arglist);
}

static void stub_silc_notify(SilcClient client, SilcClientConnection conn,
                             SilcNotifyType type, ...)
{
    va_list va;
    PyObject *arglist = NULL;
    SilcClientEntry client_entry, client_entry2;
    SilcChannelEntry channel, channel2;
    SilcServerEntry server_entry;
    char *name, *tmp;
    char buf[512];
    SilcIdType idtype;
    SilcStatus error;
    void *entry;

    va_start(va, type);

    if (callbacks[SILC_NOTIFY].callback == NULL)
        return;

    switch(type) {
    case SILC_NOTIFY_TYPE_NONE:
        /* Some generic notice from server */
        arglist = Py_BuildValue("is", SILC_NOTIFY_TYPE_NONE, (char *)va_arg(va, char *));
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_INVITE:
        /* Invited or modified invite list */
        channel = va_arg(va, SilcChannelEntry);
        name = va_arg(va, char *);
        client_entry = va_arg(va, SilcClientEntry);

        memset(buf, 0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "%s@%s",
                 client_entry->username, client_entry->hostname);

        arglist = Py_BuildValue("issss", SILC_NOTIFY_TYPE_INVITE, channel ? channel->channel_name :
                                name, client_entry->nickname, buf);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_JOIN:
        /* Joined channel */
        client_entry = va_arg(va, SilcClientEntry);
        channel = va_arg(va, SilcChannelEntry);

        memset(buf, 0, sizeof(buf));
        if (client_entry->username)
            snprintf(buf, sizeof(buf) - 1, "%s@%s",
                     client_entry->username, client_entry->hostname);

        arglist = Py_BuildValue("isss", SILC_NOTIFY_TYPE_JOIN, channel->channel_name, 
                                client_entry->nickname, buf);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_LEAVE:
        /* Left a channel */
        client_entry = va_arg(va, SilcClientEntry);
        channel = va_arg(va, SilcChannelEntry);

        memset(buf, 0, sizeof(buf));
        if (client_entry->username)
            snprintf(buf, sizeof(buf) - 1, "%s@%s",
                     client_entry->username, client_entry->hostname);

        arglist = Py_BuildValue("isss", SILC_NOTIFY_TYPE_LEAVE, channel->channel_name, 
                                client_entry->nickname, buf);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_SIGNOFF:
        /* Left the network */
        client_entry = va_arg(va, SilcClientEntry);

        if (client_entry->nickname) {
            memset(buf, 0, sizeof(buf));
            snprintf(buf, sizeof(buf) - 1, "%s@%s",
                     client_entry->username, client_entry->hostname);
            arglist = Py_BuildValue("iss", SILC_NOTIFY_TYPE_SIGNOFF, 
                                    client_entry->nickname, buf);
            PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        } else {
            arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_SIGNOFF);
            PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        }
        break;

    case SILC_NOTIFY_TYPE_TOPIC_SET:
        /* Changed topic */
        idtype = va_arg(va, int);
        entry = va_arg(va, void *);
        tmp = va_arg(va, char *);
        channel = va_arg(va, SilcChannelEntry);

        if (idtype == SILC_ID_CLIENT) {
            client_entry = (SilcClientEntry)entry;
            memset(buf, 0, sizeof(buf));
            snprintf(buf, sizeof(buf) - 1, "%s@%s",
                     client_entry->username, client_entry->hostname);
            arglist = Py_BuildValue("issss", SILC_NOTIFY_TYPE_TOPIC_SET, channel->channel_name, tmp, 
                                    client_entry->nickname, buf);
            PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        } else if (idtype == SILC_ID_SERVER) {
            server_entry = (SilcServerEntry)entry;
            arglist = Py_BuildValue("issss", SILC_NOTIFY_TYPE_TOPIC_SET, channel->channel_name, tmp, 
                                    server_entry->server_name);
            PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        } else if (idtype == SILC_ID_CHANNEL) {
            channel = (SilcChannelEntry)entry;
            arglist = Py_BuildValue("iss", SILC_NOTIFY_TYPE_TOPIC_SET, channel->channel_name, tmp);
            PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        }
        break;

    case SILC_NOTIFY_TYPE_NICK_CHANGE:
        /* Changed nickname  */
        client_entry = va_arg(va, SilcClientEntry);
        client_entry2 = va_arg(va, SilcClientEntry);

        if (!strcmp(client_entry->nickname, client_entry2->nickname))
            break;

        memset(buf, 0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "%s@%s",
                 client_entry2->username, client_entry2->hostname);
        arglist = Py_BuildValue("iss", SILC_NOTIFY_TYPE_TOPIC_SET, client_entry2->nickname, 
                                client_entry->nickname, buf);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_CMODE_CHANGE:
        /* Changed channel mode FIXME: implement this fully*/
        arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_CMODE_CHANGE);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_CUMODE_CHANGE:
        /* Changed user's mode on channel FIXME: implement this fully*/
        arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_CUMODE_CHANGE);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_MOTD:
        /* Received MOTD */
        tmp = va_arg(va, char *);
        arglist = Py_BuildValue("is", SILC_NOTIFY_TYPE_MOTD, tmp);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_KICKED:
        /* Someone was kicked from channel */
        client_entry = va_arg(va, SilcClientEntry);
        tmp = va_arg(va, char *);
        client_entry2 = va_arg(va, SilcClientEntry);
        channel = va_arg(va, SilcChannelEntry);

        arglist = Py_BuildValue("issss", SILC_NOTIFY_TYPE_KICKED, channel->channel_name,
                                client_entry->nickname, client_entry2 ? client_entry2->nickname : "",
                                tmp ? tmp : "");
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_KILLED:
        /* Someone was killed from the network */
        client_entry = va_arg(va, SilcClientEntry);
        tmp = va_arg(va, char *);
        idtype = va_arg(va, int);
        entry = va_arg(va, SilcClientEntry);


        if (idtype == SILC_ID_CLIENT) {
            client_entry2 = (SilcClientEntry)entry;
            arglist = Py_BuildValue("issss", SILC_NOTIFY_TYPE_KILLED, channel->channel_name,
                                    client_entry->nickname, client_entry2 ? client_entry2->nickname : "",
                                    tmp ? tmp : "");
        } else if (idtype == SILC_ID_SERVER) {
            server_entry = (SilcServerEntry)entry;
            arglist = Py_BuildValue("isss", SILC_NOTIFY_TYPE_KILLED, client_entry->nickname, 
                                    server_entry->server_name, tmp ? tmp : "");
        } else if (idtype == SILC_ID_CHANNEL) {
            channel = (SilcChannelEntry)entry;
            arglist = Py_BuildValue("isss", SILC_NOTIFY_TYPE_KILLED, client_entry->nickname, 
                                    channel->channel_name, tmp ? tmp : "");
        }
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_CHANNEL_CHANGE:
        arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_CHANNEL_CHANGE);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_SERVER_SIGNOFF:
        arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_CHANNEL_CHANGE);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_ERROR:
        error = va_arg(va, int);
        arglist = Py_BuildValue("iis", SILC_NOTIFY_TYPE_MOTD, error, 
                                silc_get_status_message(error));
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    case SILC_NOTIFY_TYPE_WATCH:
        arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_WATCH);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;

    default:
        arglist = Py_BuildValue("i", SILC_NOTIFY_TYPE_NONE);
        PyEval_CallObject(callbacks[SILC_NOTIFY].callback, arglist);
        break;
    }
    va_end(va);
}

static void stub_silc_command(SilcClient client, SilcClientConnection conn,
                              SilcClientCommandContext cmd_context, bool success,
                              SilcCommand command, SilcStatus status)
{
    PyObject *arglist = NULL;
    char *str;
    
    if (!success) {
        asprintf(&str, "Command %s reply: %s", silc_get_command_name(command),
                 silc_get_status_message(status));
        PyErr_SetString(SilcClientError, str);
        return;
    }

    if (callbacks[SILC_COMMAND].callback == NULL)
        return;

    arglist = Py_BuildValue("iss", command, silc_get_command_name(command),
                           silc_get_status_message(status));
    PyEval_CallObject(callbacks[SILC_COMMAND].callback, arglist);
}

static void stub_silc_command_reply(SilcClient client, SilcClientConnection conn,
                                    SilcCommandPayload cmd_payload, bool success,
                                    SilcCommand command, SilcStatus status, ...)
{
    char *str = NULL;
    PyObject *arglist = NULL;
    silcclientobject *self = client->application;
    va_list vp;
    
    if (status != SILC_STATUS_OK) {
        asprintf(&str, "Command %s reply: %s", silc_get_command_name(command),
                 silc_get_status_message(status));
        PyErr_SetString(SilcClientError, str);
        return;
    }
    
    if (callbacks[SILC_COMMAND_REPLY].callback == NULL)
        return;

    va_start(vp, status);

    switch(command) {
    case SILC_COMMAND_WHOIS:
        {
            char buf[1024], *nickname, *username, *realname, *nick;
            unsigned char *fingerprint;
            SilcUInt32 idle, mode;
            SilcBuffer channels, user_modes;
            SilcClientEntry client_entry;
            SilcDList attrs;

            if (status == SILC_STATUS_ERR_NO_SUCH_NICK) {
                unsigned char *tmp =
                    silc_argument_get_arg_type(silc_command_get_args(cmd_payload),
                                               2, NULL);
                arglist = Py_BuildValue("iiss", SILC_COMMAND_WHOIS, SILC_STATUS_ERR_NO_SUCH_NICK, tmp,
                                        silc_get_status_message(status));
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                break;
            } else if (status == SILC_STATUS_ERR_NO_SUCH_CLIENT_ID) {
                SilcUInt32 tmp_len;
                unsigned char *tmp =
                    silc_argument_get_arg_type(silc_command_get_args(cmd_payload),
                                               2, &tmp_len);
                if (tmp) {
                    SilcClientID *client_id = silc_id_payload_parse_id(tmp, tmp_len,
                                                                       NULL);
                    if (client_id) {
                        client_entry = silc_client_get_client_by_id(client, conn,
                                                                    client_id);
                        if (client_entry && client_entry->nickname)
                            arglist = Py_BuildValue("iiss", SILC_COMMAND_WHOIS, 
                                                    SILC_STATUS_ERR_NO_SUCH_CLIENT_ID,
                                                    client_entry->nickname,
                                                    silc_get_status_message(status));
                            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                        silc_free(client_id);
                    }
                }
                break;
            } else if (!success) {
                asprintf(&str, "WHOIS: %s", silc_get_status_message(status));
                PyErr_SetString(SilcClientError, str);
                return;
            }

            client_entry = va_arg(vp, SilcClientEntry);
            nickname = va_arg(vp, char *);
            username = va_arg(vp, char *);
            realname = va_arg(vp, char *);
            channels = va_arg(vp, SilcBuffer);
            mode = va_arg(vp, SilcUInt32);
            idle = va_arg(vp, SilcUInt32);
            fingerprint = va_arg(vp, unsigned char *);
            user_modes = va_arg(vp, SilcBuffer);
            attrs = va_arg(vp, SilcDList);

            silc_parse_userfqdn(nickname, &nick, NULL);
            arglist = Py_BuildValue("iisssss", SILC_COMMAND_WHOIS, status, nickname, 
                                    client_entry->username, client_entry->hostname, 
                                    nick, client_entry->nickname);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            silc_free(nick);

            if (channels && user_modes) {
                SilcUInt32 *umodes;
                SilcDList list = silc_channel_payload_parse_list(channels->data,
                                                                 channels->len);
                if (list && silc_get_mode_list(user_modes, silc_dlist_count(list),
                                               &umodes)) {
                    SilcChannelPayload entry;
                    int i = 0;

                    memset(buf, 0, sizeof(buf));
                    silc_dlist_start(list);
                    while ((entry = silc_dlist_get(list)) != SILC_LIST_END) {
                        SilcUInt32 name_len;
                        char *m = silc_client_chumode_char(umodes[i++]);
                        char *name = silc_channel_get_name(entry, &name_len);

                        if (m)
                            silc_strncat(buf, sizeof(buf) - 1, m, strlen(m));
                        silc_strncat(buf, sizeof(buf) - 1, name, name_len);
                        silc_strncat(buf, sizeof(buf) - 1, " ", 1);
                        silc_free(m);
                    }

                    arglist = Py_BuildValue("iis", SILC_COMMAND_WHOIS, status, buf);
                    PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                    silc_channel_payload_list_free(list);
                    silc_free(umodes);
                }
            }

            if (mode) {
                memset(buf, 0, sizeof(buf));
                silc_get_umode_string(mode, buf, sizeof(buf - 1));
                arglist = Py_BuildValue("iis", SILC_COMMAND_WHOIS, status, buf);
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            }

            if (idle && nickname) {
                memset(buf, 0, sizeof(buf));
                snprintf(buf, sizeof(buf) - 1, "%lu %s",
                         idle > 60 ? (idle / 60) : idle,
                         idle > 60 ? "minutes" : "seconds");

                arglist = Py_BuildValue("iis", SILC_COMMAND_WHOIS, status, buf);
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            }

            if (fingerprint) {
                fingerprint = silc_fingerprint(fingerprint, 20);
                arglist = Py_BuildValue("iis", SILC_COMMAND_WHOIS, status, fingerprint);
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                silc_free(fingerprint);
            }
        }
        break;
        
    case SILC_COMMAND_IDENTIFY:
        {
            SilcClientEntry client_entry;

            if (status == SILC_STATUS_ERR_NO_SUCH_NICK) {
                unsigned char *tmp =
                    silc_argument_get_arg_type(silc_command_get_args(cmd_payload),
                                               2, NULL);
                if (tmp) {
                    arglist = Py_BuildValue("iiss", SILC_COMMAND_IDENTIFY, SILC_STATUS_ERR_NO_SUCH_NICK, 
                                            tmp, silc_get_status_message(status));
                    PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                }
                break;
            } else if (status == SILC_STATUS_ERR_NO_SUCH_CLIENT_ID) {
                SilcUInt32 tmp_len;
                unsigned char *tmp =
                    silc_argument_get_arg_type(silc_command_get_args(cmd_payload),
                                               2, &tmp_len);
                if (tmp) {
                    SilcClientID *client_id = silc_id_payload_parse_id(tmp, tmp_len,
                                                                       NULL);
                    if (client_id) {
                        client_entry = silc_client_get_client_by_id(client, conn,
                                                                    client_id);
                        if (client_entry && client_entry->nickname) {
                            arglist = Py_BuildValue("iiss", SILC_COMMAND_IDENTIFY, 
                                                    SILC_STATUS_ERR_NO_SUCH_CLIENT_ID, 
                                                    client_entry->nickname,
                                                    silc_get_status_message(status));
                            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                        }
                        silc_free(client_id);
                    }
                }
                break;
            }

            break;
        }
    case SILC_COMMAND_WHOWAS:
        {
            char *nickname, *username, *realname;

            if (status == SILC_STATUS_ERR_NO_SUCH_NICK) {
                char *tmp =
                    silc_argument_get_arg_type(silc_command_get_args(cmd_payload),
                                               2, NULL);
                if (tmp) {
                    arglist = Py_BuildValue("iiss", SILC_COMMAND_WHOWAS, SILC_STATUS_ERR_NO_SUCH_NICK, 
                                            tmp, silc_get_status_message(status));
                    PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                }
                break;
            } else if (!success) {
                    arglist = Py_BuildValue("iis", SILC_COMMAND_WHOWAS, status,
                                            silc_get_status_message(status));
                    PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                return;
            }

            (void)va_arg(vp, SilcClientEntry);
            nickname = va_arg(vp, char *);
            username = va_arg(vp, char *);
            realname = va_arg(vp, char *);
            
            arglist = Py_BuildValue("iisss", SILC_COMMAND_WHOWAS, status, nickname, username, 
                                    realname ? realname : "");
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
        }
        break;

    case SILC_COMMAND_INVITE:
        {
            SilcChannelEntry channel;
            SilcBuffer payload;
            SilcArgumentPayload invite_list;
            SilcUInt16 argc;

            if (!success)
                return;

            channel = va_arg(vp, SilcChannelEntry);
            payload = va_arg(vp, SilcBuffer);

            arglist = Py_BuildValue("iis", SILC_COMMAND_INVITE, status, channel->channel_name);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            if (!payload) 
                break;
                
            SILC_GET16_MSB(argc, payload->data);
            invite_list = silc_argument_payload_parse(payload->data + 2,
                                                      payload->len - 2, argc);
            if (!invite_list);
                break;

        }
        break;

    case SILC_COMMAND_JOIN:
        {
            char *channel, *mode, *topic;
            char *ret_channel;
            SilcChannelEntry ret_channel_entry;
            SilcUInt32 modei;
            SilcChannelEntry channel_entry;
            SilcBuffer client_id_list;
            SilcUInt32 list_count;

            if (!success)
                return;

            channel = va_arg(vp, char *);
            channel_entry = va_arg(vp, SilcChannelEntry);
            modei = va_arg(vp, SilcUInt32);
            (void)va_arg(vp, SilcUInt32);
            (void)va_arg(vp, unsigned char *);
            (void)va_arg(vp, unsigned char *);
            (void)va_arg(vp, unsigned char *);
            topic = va_arg(vp, char *);
            (void)va_arg(vp, unsigned char *);
            list_count = va_arg(vp, SilcUInt32);
            client_id_list = va_arg(vp, SilcBuffer);

            if (channel && !silc_hash_table_find(self->channels, channel, NULL, NULL))
                silc_hash_table_add(self->channels, (char *)strdup(channel), channel_entry);

            arglist = Py_BuildValue("is", SILC_COMMAND_JOIN, channel);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            break;
        }

    case SILC_COMMAND_NICK:
        {
            SilcClientEntry client_entry = va_arg(vp, SilcClientEntry);

            if (!success)
                return;

            arglist = Py_BuildValue("is", SILC_COMMAND_NICK, client_entry->nickname);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            break;
        }

    case SILC_COMMAND_LIST:
        {
            char *topic, *name;
            int usercount;
            char users[20];
            char tmp[256], *cp, *dm = NULL;

            if (!success)
                return;

            (void)va_arg(vp, SilcChannelEntry);
            name = va_arg(vp, char *);
            if (!name)
                return;
            topic = va_arg(vp, char *);
            usercount = va_arg(vp, int);

            if (topic && silc_utf8_valid(topic, strlen(topic))) {
                memset(tmp, 0, sizeof(tmp));
                cp = tmp;
                if (strlen(topic) > sizeof(tmp) - 1) {
                    dm = silc_calloc(strlen(topic) + 1, sizeof(*dm));
                    cp = dm;
                }

                silc_utf8_decode(topic, strlen(topic), SILC_STRING_LOCALE,
                                 cp, strlen(topic));
                topic = cp;
            }

            if (!usercount)
                snprintf(users, sizeof(users) - 1, "N/A");
            else
                snprintf(users, sizeof(users) - 1, "%d", usercount);
            arglist = Py_BuildValue("iisss", SILC_COMMAND_LIST, status,
                                    name, users, topic ? topic : "");
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            silc_free(dm);
        }
        break;

    case SILC_COMMAND_UMODE:
        {
            SilcUInt32 mode;

            if (!success)
                return;

            mode = va_arg(vp, SilcUInt32);

            arglist = Py_BuildValue("ii", SILC_COMMAND_UMODE, mode);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
        }
        break;

    case SILC_COMMAND_OPER:
        if (!success)
            return;

        arglist = Py_BuildValue("i", SILC_COMMAND_OPER);
        PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
        break;

    case SILC_COMMAND_SILCOPER:
        if (!success)
            return;

        arglist = Py_BuildValue("i", SILC_COMMAND_SILCOPER);
        PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
        break;

    case SILC_COMMAND_USERS:
        {
            SilcHashTableList htl;
            SilcChannelEntry channel;
            SilcChannelUser chu;

            if (!success)
                return;

            channel = va_arg(vp, SilcChannelEntry);

            silc_hash_table_list(channel->user_list, &htl);
            while (silc_hash_table_get(&htl, NULL, (void *)&chu)) {
                SilcClientEntry e = chu->client;
                char stat[5], *mode;

                if (!e->nickname)
                    continue;

                memset(stat, 0, sizeof(stat));
                mode = silc_client_chumode_char(chu->mode);
                if (e->mode & SILC_UMODE_GONE)
                    strcat(stat, "G");
                else if (e->mode & SILC_UMODE_INDISPOSED)
                    strcat(stat, "I");
                else if (e->mode & SILC_UMODE_BUSY)
                    strcat(stat, "B");
                else if (e->mode & SILC_UMODE_PAGE)
                    strcat(stat, "P");
                else if (e->mode & SILC_UMODE_HYPER)
                    strcat(stat, "H");
                else if (e->mode & SILC_UMODE_ROBOT)
                    strcat(stat, "R");
                else if (e->mode & SILC_UMODE_ANONYMOUS)
                    strcat(stat, "?");
                else
                    strcat(stat, "A");
                if (mode)
                    strcat(stat, mode);

                arglist = Py_BuildValue("isssss", SILC_COMMAND_USERS, channel->channel_name, 
                                        e->nickname, stat,
                                        e->username ? e->username : "",
                                        e->hostname ? e->hostname : "",
                                        e->realname ? e->realname : "");
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                if (mode)
                    silc_free(mode);
            }
            silc_hash_table_list_reset(&htl);
        }
        break;

    case SILC_COMMAND_BAN:
        {
            SilcChannelEntry channel;
            SilcBuffer payload;
            SilcArgumentPayload ban_list;
            SilcUInt16 argc;

            if (!success)
                return;

            channel = va_arg(vp, SilcChannelEntry);
            payload = va_arg(vp, SilcBuffer);

            if (payload) {
                SILC_GET16_MSB(argc, payload->data);
                ban_list = silc_argument_payload_parse(payload->data + 2,
                                                       payload->len - 2, argc);
                if (ban_list) {
                    arglist = Py_BuildValue("iss", SILC_COMMAND_BAN, channel->channel_name, ban_list);
                    PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                    
                    silc_argument_payload_free(ban_list);
                }
            }
        }
        break;

    case SILC_COMMAND_GETKEY:
        {
            SilcIdType id_type;
            void *entry;
            SilcPublicKey public_key;
            unsigned char *pk;
            SilcUInt32 pk_len;
            char *name;

            if (!success)
                return;

            id_type = va_arg(vp, SilcUInt32);
            entry = va_arg(vp, void *);
            public_key = va_arg(vp, SilcPublicKey);

            if (public_key) {
                pk = silc_pkcs_public_key_encode(public_key, &pk_len);


                name = (id_type == SILC_ID_CLIENT ?
                        ((SilcClientEntry)entry)->nickname :
                        ((SilcServerEntry)entry)->server_name);
                
                arglist = Py_BuildValue("iiss#s", SILC_COMMAND_GETKEY, id_type,
                                        name, pk, pk_len, silc_hash_fingerprint(NULL, pk, pk_len));
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
                silc_free(pk);
            } 
        }
        break;

    case SILC_COMMAND_INFO:
        {
            SilcServerEntry server_entry;
            char *server_name;
            char *server_info;

            if (!success)
                return;

            server_entry = va_arg(vp, SilcServerEntry);
            server_name = va_arg(vp, char *);
            server_info = va_arg(vp, char *);

            if (server_name && server_info )
            {
                arglist = Py_BuildValue("iss", SILC_COMMAND_INFO, server_name,
                                        server_info);
                PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            }
        }
        break;

    case SILC_COMMAND_TOPIC:
        {
            SilcChannelEntry channel;
            char *topic;
            char tmp[256], *cp, *dm = NULL;

            if (!success)
                return;

            channel = va_arg(vp, SilcChannelEntry);
            topic = va_arg(vp, char *);

            if (topic && silc_utf8_valid(topic, strlen(topic))) {
                memset(tmp, 0, sizeof(tmp));
                cp = tmp;
                if (strlen(topic) > sizeof(tmp) - 1) {
                    dm = silc_calloc(strlen(topic) + 1, sizeof(*dm));
                    cp = dm;
                }

                silc_utf8_decode(topic, strlen(topic), SILC_STRING_LOCALE,
                                 cp, strlen(topic));
                topic = cp;
            }

            arglist = Py_BuildValue("iss", SILC_COMMAND_TOPIC, channel->channel_name,
                                    topic);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
            silc_free(dm);
        }
        break;

    case SILC_COMMAND_WATCH:
        break;

    case SILC_COMMAND_STATS:
        {
            SilcUInt32 starttime, uptime, my_clients, my_channels, my_server_ops,
                       my_router_ops, cell_clients, cell_channels, cell_servers,
                       clients, channels, servers, routers, server_ops, router_ops;
            SilcUInt32 buf_len;
            SilcBufferStruct buf;
            unsigned char *tmp_buf;
            char tmp[40];
            const char *tmptime;
            int days, hours, mins, secs;

            if (!success)
                return;

            tmp_buf = va_arg(vp, unsigned char *);
            buf_len = va_arg(vp, SilcUInt32);

            if (!tmp_buf || !buf_len) {
                return;
            }

            silc_buffer_set(&buf, tmp_buf, buf_len);
            silc_buffer_unformat(&buf,
                                 SILC_STR_UI_INT(&starttime),
                                 SILC_STR_UI_INT(&uptime),
                                 SILC_STR_UI_INT(&my_clients),
                                 SILC_STR_UI_INT(&my_channels),
                                 SILC_STR_UI_INT(&my_server_ops),
                                 SILC_STR_UI_INT(&my_router_ops),
                                 SILC_STR_UI_INT(&cell_clients),
                                 SILC_STR_UI_INT(&cell_channels),
                                 SILC_STR_UI_INT(&cell_servers),
                                 SILC_STR_UI_INT(&clients),
                                 SILC_STR_UI_INT(&channels),
                                 SILC_STR_UI_INT(&servers),
                                 SILC_STR_UI_INT(&routers),
                                 SILC_STR_UI_INT(&server_ops),
                                 SILC_STR_UI_INT(&router_ops),
                                 SILC_STR_END);

            tmptime = silc_get_time(starttime);
            
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local server start time", tmptime);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            days = uptime / (24 * 60 * 60);
            uptime -= days * (24 * 60 * 60);
            hours = uptime / (60 * 60);
            uptime -= hours * (60 * 60);
            mins = uptime / 60;
            uptime -= mins * 60;
            secs = uptime;
            snprintf(tmp, sizeof(tmp) - 1, "%d days %d hours %d mins %d secs",
                     days, hours, mins, secs);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local server uptime", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)my_clients);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local server clients", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)my_channels);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local server channels", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)my_server_ops);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local server operators", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)my_router_ops);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local router operators", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)cell_clients);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local cell clients", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)cell_channels);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local cell channels", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)cell_servers);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Local cell servers", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)clients);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Total clients", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)channels);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Total channels", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)servers);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Total servers", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)routers);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Total routers", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)server_ops);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Total servers", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);

            snprintf(tmp, sizeof(tmp) - 1, "%d", (int)router_ops);
            arglist = Py_BuildValue("iss", SILC_COMMAND_STATS, "Total router operators", tmp);
            PyEval_CallObject(callbacks[SILC_COMMAND_REPLY].callback, arglist);
        }
        break;

    case SILC_COMMAND_CMODE:
        {
            SilcChannelEntry channel_entry;
            SilcBuffer channel_pubkeys;

            channel_entry = va_arg(vp, SilcChannelEntry);
            (void)va_arg(vp, SilcUInt32);
            (void)va_arg(vp, SilcPublicKey);
            channel_pubkeys = va_arg(vp, SilcBuffer);

            if (!success || !channel_entry || !channel_entry->channel_name)
                return;

            /* FIXME: Complete_ */

        }
        break;

    case SILC_COMMAND_LEAVE:
        {
        }
        break;

    }

    va_end(vp);
}

static void stub_silc_connected(SilcClient client, SilcClientConnection conn,
                                SilcClientConnectionStatus status)
{
    PyObject *arglist = Py_BuildValue("(i)", status);
    silcclientobject *self = client->application;
    char *str;
    
    if (status != SILC_CLIENT_CONN_SUCCESS && status != SILC_CLIENT_CONN_SUCCESS_RESUME) {
        silc_client_close_connection(client, conn);
        asprintf(&str, "Could not connect to server: %d", status);
        PyErr_SetString(SilcClientError, str);
        return;
    }
    self->conn = conn;
    self->connected = TRUE;
    
    if (callbacks[SILC_CONNECTED].callback != NULL)
        PyEval_CallObject(callbacks[SILC_CONNECTED].callback, arglist);
}

static void stub_silc_disconnected(SilcClient client, SilcClientConnection conn,
                                   SilcStatus status, const char *message)
{
    PyObject *arglist = Py_BuildValue("(is)", status, message);
    silcclientobject *self = client->application;

    self->connected = FALSE;
    if (callbacks[SILC_DISCONNECTED].callback != NULL)
        PyEval_CallObject(callbacks[SILC_DISCONNECTED].callback, arglist);
    
}

typedef struct {
    SilcGetAuthMeth completion;
    void *context;
} *InternalGetAuthMethod;

/* FIXME: add password method */
void silc_get_auth_method_callback(SilcClient client,
                                   SilcClientConnection conn,
                                   SilcAuthMethod auth_meth,
                                   void *context)
{
    InternalGetAuthMethod internal = (InternalGetAuthMethod)context;
    
    switch (auth_meth) {
    case SILC_AUTH_NONE:
        (*internal->completion)(TRUE, auth_meth, NULL, 0, internal->context);
        break;
    case SILC_AUTH_PUBLIC_KEY:
        (*internal->completion)(TRUE, auth_meth, NULL, 0, internal->context);
        break;
    }
    silc_free(internal);
}

static void stub_silc_get_auth_method(SilcClient client, SilcClientConnection conn,
                                      char *hostname, SilcUInt16 port,
                                      SilcGetAuthMeth completion,
                                      void *context)
{
    InternalGetAuthMethod internal;

    internal = silc_calloc(1, sizeof(*internal));
    internal->completion = completion;
    internal->context = context;

    silc_client_request_authentication_method(client, conn,
                                              silc_get_auth_method_callback,
                                              internal);

}

static void stub_silc_verify_public_key(SilcClient client, SilcClientConnection conn,
                                        SilcSocketType conn_type, unsigned char *pk,
                                        SilcUInt32 pk_len, SilcSKEPKType pk_type,
                                        SilcVerifyPublicKey completion, void *context)
{
    /* FIXME: complete this */
    completion(TRUE, context);
}

static void stub_silc_ask_passphrase(SilcClient client, SilcClientConnection conn,
                                     SilcAskPassphrase completion, void *context)
{
    PyObject *arglist = Py_BuildValue("(s)", "Foo");
    if (callbacks[SILC_ASK_PASSPHRASE].callback != NULL)
        PyEval_CallObject(callbacks[SILC_ASK_PASSPHRASE].callback, arglist);
}

static void stub_silc_failure(SilcClient client, SilcClientConnection conn,
                              SilcProtocol protocol, void *failure)
{
    PyObject *arglist = NULL;
    char *msg = NULL;
    SilcUInt32 status;
    

    if (protocol->protocol->type == SILC_PROTOCOL_CLIENT_KEY_EXCHANGE) {
        SilcSKEStatus ske_status = (SilcSKEStatus)failure;

        status = SILC_STATUS_ERR_KEY_EXCHANGE_FAILED;
        msg = strdup(silc_ske_map_status(ske_status));
    } else if (protocol->protocol->type == SILC_PROTOCOL_CLIENT_CONNECTION_AUTH) {
        status = (SilcUInt32)failure;
        msg = strdup(silc_get_status_message((SilcStatus)status));
    } else {
        status = -1;
        asprintf(&msg, "Unknown failure");
    }
    arglist = Py_BuildValue("(is)", status, msg);
    if (callbacks[SILC_FAILURE].callback != NULL)
        PyEval_CallObject(callbacks[SILC_FAILURE].callback, arglist);
}

static bool stub_silc_key_agreement(SilcClient client, SilcClientConnection conn,
                                    SilcClientEntry client_entry, const char *hostname,
                                    SilcUInt16 port, SilcKeyAgreementCallback *completion,
                                    void **context)
{
    return FALSE;
}

static void stub_silc_ftp(SilcClient client, SilcClientConnection conn,
                          SilcClientEntry client_entry, SilcUInt32 session_id,
                          const char *hostname, SilcUInt16 port)
{
}

static void stub_silc_detach(SilcClient client, SilcClientConnection conn,
                             const unsigned char *detach_data, SilcUInt32 detach_data_len)
{
    PyObject *arglist = Py_BuildValue("(s#)", detach_data, detach_data_len);

    if (callbacks[SILC_DETACH].callback != NULL)
        PyEval_CallObject(callbacks[SILC_DETACH].callback, arglist);
}

SilcClientOperations ops = {
    stub_silc_say,                
    stub_silc_channel_message,
    stub_silc_private_message,
    stub_silc_notify,
    stub_silc_command,
    stub_silc_command_reply,
    stub_silc_connected,
    stub_silc_disconnected,
    stub_silc_get_auth_method,
    stub_silc_verify_public_key,
    stub_silc_ask_passphrase,
    stub_silc_failure,
    stub_silc_key_agreement,
    stub_silc_ftp,
    stub_silc_detach
};

static PyObject *
silcclient_callbacks(PyObject *self, PyObject *args)
{
    PyObject *tmp = NULL;
    PyObject *p   = NULL;
    int i;

    if (!PyArg_ParseTuple(args, "O", &p)) {
        PyErr_SetString(PyExc_TypeError, "Failed to parse args");
        return NULL;
    }

    if (!PyDict_Check(p)) {
        PyErr_SetString(PyExc_TypeError, "Expected a dictionary");
        return NULL;
    }

    for (i = 0; callbacks[i].name != NULL; i++) {
        tmp = PyDict_GetItemString(p, callbacks[i].name);
        /* FIXME: uncomment the following to force user-defined subroutines for every signal */
        /*
        if (!PyCallable_Check(tmp)) {
            asprintf(str, "Expected a callable for %s", callbacks[i].name);
            PyErr_SetString(SilcClientError, str);
            return NULL;
        }
        */
        if (PyCallable_Check(tmp))
            callbacks[i].callback = tmp;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
silcclient_loadkeys(silcclientobject *self, PyObject *args)
{
    char *pubkey_file = NULL;
    char *prvkey_file = NULL;
    char *password    = NULL;
    char *str;
    int ret;
    
    if (!PyArg_ParseTuple(args, "sss", &pubkey_file, &prvkey_file, &password)) {
        PyErr_SetString(PyExc_TypeError, "Failed to parse args");
        return NULL;
    }

    if (silc_load_key_pair(pubkey_file, prvkey_file, password,
                            &self->client->pkcs,
                            &self->client->public_key,
                            &self->client->private_key) == FALSE) {
        ret = silc_load_key_pair(pubkey_file, prvkey_file, NULL,
                                &self->client->pkcs,
                                &self->client->public_key,
                                &self->client->private_key);
        if (ret == FALSE) {
            asprintf(&str, "Failed to load keys : %d", ret);
            PyErr_SetString(SilcClientError, str);
            return NULL;
        }
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
silcclient_createkeys(silcclientobject *self, PyObject *args, PyObject *keywds)
{
    int key_length       = 2048;
    int ret;
    char *str;
    char *pkcs_name      = "rsa";
    char *pubkey_file    = NULL;
    char *prvkey_file    = NULL;
    char *password       = NULL;
    char *pub_identifier = NULL;
    static char *kwlist[] = {"pubkey_file", "prvkey_file", "password", "identifier", 
        "pkcs", "pkcs_len", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "sss|ssi", kwlist,
                                     &pubkey_file, &prvkey_file, &password, 
                                     &pub_identifier, &pkcs_name, &key_length)) {
        PyErr_SetString(PyExc_TypeError, "Failed to parse args");
        return NULL;
    }

    if ((pub_identifier != NULL) && (strlen(pub_identifier) == 0))
        pub_identifier = NULL;

    ret = silc_create_key_pair(pkcs_name, key_length, pubkey_file, prvkey_file, 
                               pub_identifier, password, 
                               &self->client->pkcs,
                               &self->client->public_key,
                               &self->client->private_key, FALSE);
    if (!ret) {
        asprintf(&str, "Failed to create keys : %s", silc_get_status_message(ret));
        PyErr_SetString(SilcClientError, str);
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *silcclient_connect(silcclientobject *self, PyObject *args, PyObject *keywds)
{
    char *server;
    int port = 706;
    static char *kwlist[] = {"server", "port", NULL};
    
    if (!PyArg_ParseTupleAndKeywords(args, keywds, "s|i", kwlist, 
                                     &server, &port)) {
        PyErr_SetString(PyExc_TypeError, "Failed to parse args");
        return NULL;
    }
    silc_client_connect_to_server(self->client, NULL, port,
                                  server, self);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *silcclient_command(silcclientobject *self, PyObject *args)
{
    char *command;
    if (self->connected == FALSE) {
        PyErr_SetString(SilcClientError, "Not connected to a SILC server");
        return NULL;
    }
    
    if (!PyArg_ParseTuple(args, "s", &command)) {
        PyErr_SetString(PyExc_TypeError, "Failed to parse args");
        return NULL;
    }

    if (!silc_client_command_call(self->client, self->conn, command)) {
        PyErr_SetString(SilcClientError, "Command not known");
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *silcclient_send_channel(silcclientobject *self, PyObject *args, PyObject *keywds)
{
    char *message = NULL;
    char *channel = NULL;
    char *str;
    SilcChannelEntry channel_entry;
    int sign = TRUE;
    static char *kwlist[] = {"channel", "message", "signed", NULL};
    
    if (self->connected == FALSE) {
        PyErr_SetString(SilcClientError, "Not connected to a SILC server");
        return NULL;
    }
    
    if (!PyArg_ParseTupleAndKeywords(args, keywds, "ss|i", kwlist, 
                                     &channel, &message, &sign)) {
        PyErr_SetString(PyExc_TypeError, "Failed to parse args");
        return NULL;
    }

    if (silc_hash_table_find(self->channels, channel, NULL, (void *)&channel_entry) == FALSE) {
        asprintf(&str, "Not on channel : %s", channel);
        PyErr_SetString(SilcClientError, str);
        return NULL;
    }
    
    if (!silc_client_send_channel_message(self->client, self->conn, channel_entry, NULL, 
                                          sign ? SILC_MESSAGE_FLAG_SIGNED : 0, message, 
                                          strlen(message), FALSE)) {
        asprintf(&str, "Could not send message %s on channel %s", message, channel);
        PyErr_SetString(SilcClientError, str);
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *silcclient_run(silcclientobject *self, PyObject *args)
{
    silc_client_run(self->client);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *silcclient_run_one(silcclientobject *self, PyObject *args)
{
    silc_client_run_one(self->client);
    Py_INCREF(Py_None);
    return Py_None;
}

/* Instance method table */
static struct PyMethodDef silcclient_methods[] = {
    { "setcallbacks",  (PyCFunction)silcclient_callbacks,    METH_VARARGS},
    { "loadkeys",      (PyCFunction)silcclient_loadkeys,     METH_VARARGS},
    { "createkeys",    (PyCFunction)silcclient_createkeys,   METH_VARARGS | METH_KEYWORDS},
    { "run",           (PyCFunction)silcclient_run,          METH_VARARGS},
    { "runone",        (PyCFunction)silcclient_run_one,      METH_VARARGS},
    { "connect",       (PyCFunction)silcclient_connect,      METH_VARARGS | METH_KEYWORDS},
    { "command",       (PyCFunction)silcclient_command,      METH_VARARGS | METH_KEYWORDS},
    { "send_channel",  (PyCFunction)silcclient_send_channel, METH_VARARGS | METH_KEYWORDS},
    { NULL, NULL }
};

/* Basic operations */
static silcclientobject * new_silcclient()
{
    silcclientobject *self;
    self = PyObject_NEW(silcclientobject, &SilcClientObjectType);
    if (self == NULL)
        return NULL;

    self->client = silc_client_alloc(&ops, NULL, self, NULL);
    self->client->username = silc_get_username();
    self->client->hostname = silc_net_localhost();
    self->client->realname = silc_get_username();

    if (!silc_client_init(self->client)) {
        PyErr_SetString(SilcClientError, "Failed to init");
        return NULL;
    }

    self->channels = silc_hash_table_alloc(0, silc_hash_string, NULL, hash_compare_str, NULL,
                                           NULL, NULL, TRUE);
    self->connected = FALSE;
    return self;
}	

static void
dealloc_silcclient(silcclientobject *self) 
{
    silc_client_free(self->client);
    silc_hash_table_free(self->channels);
    PyMem_DEL(self);
}

static PyObject *silcclient_getattr(silcclientobject *self, char *attrname)
{
    if (strcmp(attrname, "username") == 0) 
        return Py_BuildValue("s", self->client->username);
    else if (strcmp(attrname, "hostname") == 0)
        return Py_BuildValue("s", self->client->hostname);
    else if (strcmp(attrname, "realname") == 0)
        return Py_BuildValue("s", self->client->realname);

    return Py_FindMethod(silcclient_methods, (PyObject *)self, attrname);
}

static PyObject *silcclient_setattr(silcclientobject *self, char *attrname, PyObject *value)
{ 
    char *newval;

    if (value == NULL)
        return NULL;

    if (strcmp(attrname, "username") == 0) {
        if (self->client->username != NULL)
            free(self->client->username);
        newval = PyString_AsString(value);
        self->client->username = newval;
    } else if (strcmp(attrname, "hostname") == 0) {
        if (self->client->hostname != NULL)
            free(self->client->hostname);
        newval = PyString_AsString(value);
        self->client->hostname = newval;
    } else if (strcmp(attrname, "realname") == 0) {
        if (self->client->realname != NULL)
            free(self->client->realname);
        newval = PyString_AsString(value);
        self->client->realname = newval;
    } else
        PyErr_SetString(PyExc_AttributeError, attrname);

    return 0;
}

static PyObject *silcclient_repr(silcclientobject *self)
{
    char rbuffer[256];
    sprintf(rbuffer, "silcclient");
    return PyString_FromString(rbuffer);
}

static PyTypeObject SilcClientObjectType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,	                                /* ob_size */
    "Client",                           /* tp_name */
    sizeof(silcclientobject),           /* tp_basicsize */
    0,                                  /* tp_itemsize */
    
    (destructor)    dealloc_silcclient, /* tp_dealloc (refcount = 0) */
    (printfunc)     0,                  /* tp_print */
    (getattrfunc)	silcclient_getattr, /* tp_getattr (x.attr) */
    (setattrfunc)   silcclient_setattr, /* tp_setattr */
    (cmpfunc)       0,                  /* tp_compare */
    (reprfunc)      silcclient_repr,    /* tp_repr (repr(x) */
    
    0,                                  /* tp_as_number */
    0,                                  /* tp_as_sequence */
    0,                                  /* tp_as_mapping */
    (hashfunc)      0,                  /* tp_hash (dict[x])*/
    (ternaryfunc)   0,                  /* tp_call x() */
    (reprfunc)      0,                  /* tp_str str(x) */
};

static PyObject *silcclientobject_new(PyObject *self, PyObject *args)
{
    return (PyObject *) new_silcclient(0);
}

static struct PyMethodDef silcclienttype_methods[] = {
    { "Client", silcclientobject_new, METH_VARARGS },
    { NULL, NULL }
};

void initclient() {
    PyObject *m, *d;

    m = Py_InitModule("client", silcclienttype_methods);
    d = PyModule_GetDict(m);
    SilcClientError = PyErr_NewException("silcclient.error", NULL, NULL);
    PyDict_SetItemString(d, "error", SilcClientError);
}
